"""
Script to automatically add potential trees in a given region, along roads and paths.

Road information is downloaded from OpenStreetMap.
Existing trees can be imported from shapefiles if desired

Trees are exported in a CSV table, a PNG diagram and an HTML interactive map.
"""
import pickle
from pathlib import Path
from collections import namedtuple

import folium
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
import numpy as np
import overpy
from pyproj import Transformer
from shapely import LineString, geometry, wkt
from shapely.ops import transform
import pandas as pd
import geopandas as gpd

from tree import Forest
from import_existing_trees import get_existing_forest

# TODO: Use Args
# TODO: Document
# TODO: Write issue
# TODO: Write tests?

# From RegionChooser, or https://transfer.hft-stuttgart.de/gitlab/circulargreensimcity/circulargreensimcity/-/wikis/Fallstudien/Gromb%C3%BChl
WKT = "POLYGON((9.947021 49.803063, 9.947011 49.800917, 9.955025 49.800810, 9.955110 49.803019, 9.947021 49.803063))"
# Replace with None if no existing tree should be imported
EXISTING_TREES = 'existing_trees/Trees_ideal_2_20240227.shp'
# EXISTING_TREES = 'existing_trees/baumkataster/Baum.shp'
EPSG_ID = 25832
# Trees will be planted every TREE_DISTANCE along roads:
TREE_DISTANCE = 10  # [m]
# Unless there's already another tree closer than MIN_DISTANCE away:
MIN_DISTANCE = TREE_DISTANCE * 0.5  # [m]
# For display purposes only:
GRID = 100  # [m]

IGNORE_ROADS = set(['primary', 'unclassified', 'secondary',
                    'secondary_link', 'trunk', 'trunk_link', 'primary_link'])


SCRIPT_DIR = Path(__file__).resolve().parent
OUTPUT_DIR = SCRIPT_DIR / 'output'
Bounds = namedtuple("Bounds", "W S E N")


def load_region(wkt_polygon):
    region = wkt.loads(wkt_polygon)
    bounds = Bounds(*region.bounds)
    return region, bounds


def get_basename(bounds):
    return f'{bounds.S}__{bounds.N}__{bounds.W}__{bounds.E}_{TREE_DISTANCE}m'.replace('.', '_')


def get_osm_roads(bounds):
    cache_dir = SCRIPT_DIR / 'cache'
    cache_dir.mkdir(exist_ok=True)

    cache_file = (cache_dir / get_basename(bounds)).with_suffix('.pickle')

    if cache_file.exists():
        print("Cache has been found. Parsing...")
        with open(cache_file, 'rb') as cache:
            ways = pickle.load(cache)
    else:
        print("Downloading data...")
        # TODO: Could add trees from OSM too.
        api = overpy.Overpass()
        result = api.query(f"""
        way({bounds.S},{bounds.W},{bounds.N},{bounds.E}) ["highway"];
        (._;>;);
        out body;
        """)
        ways = result.ways
        print("Caching data...")
        with open(cache_file, 'wb') as cache:
            pickle.dump(ways, cache)
    return ways


def set_plot(bounds, to_local_coordinates):
    x_min, y_min = to_local_coordinates.transform(bounds.W, bounds.S)
    x_max, y_max = to_local_coordinates.transform(bounds.E, bounds.N)
    ax = plt.axes()
    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)
    ax.set_aspect(1)

    x_grid = plticker.MultipleLocator(base=GRID)
    y_grid = plticker.MultipleLocator(base=GRID)
    ax.xaxis.set_major_locator(x_grid)
    ax.yaxis.set_major_locator(y_grid)
    return ax


def place_trees(forest, ways, region, to_local, tree_distance, min_distance_2) -> Forest:
    local_region = transform(to_local.transform, region)

    for way in ways:
        width = float(way.tags.get("width", 0))
        highway = way.tags.get("highway")
        if highway in IGNORE_ROADS:
            color = 'orange'
        else:
            color = 'gray'

        road_xy_s = [to_local.transform(node.lon, node.lat) for node in way.nodes]

        tree_path = LineString(road_xy_s)

        if width:
            road_as_polygon = tree_path.buffer(width / 2, cap_style='flat')
            tree_path = road_as_polygon.exterior
            displayed_width = width
        else:
            # NOTE: Could try to guess width depending on highway type.
            displayed_width = 1

        road_xs, road_ys = zip(*road_xy_s)
        plt.plot(road_xs, road_ys, linewidth=displayed_width, c=color, alpha=0.8, zorder=-1)

        distances = np.arange(0, tree_path.length, tree_distance)
        potential_trees = [tree_path.interpolate(
            distance) for distance in distances]
        if tree_path.boundary:
            potential_trees += [tree_path.boundary.geoms[-1]]

        for potential_tree in potential_trees:
            x = potential_tree.x
            y = potential_tree.y
            if local_region.contains(geometry.Point(x, y)):
                forest.add_tree_if_possible(min_distance_2, x, y,
                                            color='#DFFF00',
                                            type='Fake Tree',
                                            description='Tilia tomentosa',
                                            diameter=6,
                                            height=10,
                                            source='add_trees.py'
                                            )

    return forest


def plot_trees(bounds: Bounds, forest: Forest, tree_distance: float) -> None:
    print("Exporting diagram...")
    tree_xs, tree_ys, colors = forest.xs_ys_cs
    plt.scatter(tree_xs, tree_ys, s=2, c=colors)

    plt.grid(True)
    plt.title(f"{bounds}\nTree distance : {tree_distance} m")
    plt.gcf().set_size_inches(15, 10)
    plt.savefig(
        OUTPUT_DIR / f"{get_basename(bounds)}.png", bbox_inches='tight', dpi=300)
    print("  DONE!")


def export_map(bounds, forest, epsg_id) -> None:
    print("Exporting Map...")
    to_wgs84 = Transformer.from_crs(f"EPSG:{epsg_id}", "EPSG:4326", always_xy=True)
    interactive_map = folium.Map()
    interactive_map.fit_bounds([(bounds.S, bounds.W), (bounds.N, bounds.E)])

    for tree in forest:
        lon, lat = to_wgs84.transform(tree.x, tree.y)
        folium.Circle(
            location=[lat, lon],
            radius=tree.radius,  # [m]
            color="black",
            weight=1,
            fill_opacity=0.9,
            opacity=1,
            fill_color=tree.color,
            fill=False,  # gets overridden by fill_color
            popup=repr(tree),
            tooltip=str(tree),
        ).add_to(interactive_map)

    folium.TileLayer(
        tiles='https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}',
        attr='Esri',
        name='Esri Satellite',
        overlay=False,
        control=True
    ).add_to(interactive_map)

    interactive_map.save(OUTPUT_DIR / f"{get_basename(bounds)}_trees.html")
    print("  DONE!")


def export_csv(bounds, forest, wkt_polygon, tree_distance, min_distance, epsg_id) -> None:
    print("Exporting CSV...")
    with open(OUTPUT_DIR / f"{get_basename(bounds)}_trees.csv", "w") as csv:
        csv.write(f"# Fake trees for; {wkt_polygon}\n")
        csv.write(f"# Tree distance along roads; {tree_distance}; [m]\n")
        csv.write(f"# Minimum allowed distance between trees; {min_distance}; [m]\n")
        csv.write(f"# EPSG; {epsg_id}\n")
        csv.write("# X; Y; Type; Description; Radius; Height; Source\n")
        csv.write("# [m]; [m]; [-]; [-]; [m]; [m]; [-]\n")
        for tree in forest:
            csv.write(f"{tree.x};{tree.y};{tree.type};{tree.description};{tree.radius};{tree.height};{tree.source}\n")

    print("  DONE!")


def export_shapefile(bounds: Bounds, forest: Forest, epsg_id: str) -> None:
    print("Exporting shapefile")

    data = [{
        'x': t.x, 'y': t.y,
        'Bezeichnun': t.description, 'Baumart': t.type,
        'Baumhöhe': t.height, 'Kronenbrei': t.diameter,
        'Stammumfan': t.trunk_diameter, 'Quelle': t.source
    }
        for t in forest]

    df = pd.DataFrame.from_dict(data)
    gdf = gpd.GeoDataFrame(
        df.drop(columns=['x', 'y']),
        geometry=gpd.points_from_xy(df.x, df.y), crs=epsg_id
    )

    print(gdf)

    basename = get_basename(bounds)
    shp_dir = OUTPUT_DIR / basename
    shp_dir.mkdir(exist_ok=True)
    gdf.to_file(shp_dir / "trees.shp", encoding='UTF-8')

    print("  DONE!")


def main(wkt_polygon, epsg_id, tree_distance, min_distance, import_tree_shp) -> None:
    region, bounds = load_region(wkt_polygon)
    ways = get_osm_roads(bounds)

    if import_tree_shp:
        existing_forest = get_existing_forest(import_tree_shp)
    else:
        existing_forest = Forest()

    print(existing_forest)

    to_local = Transformer.from_crs("EPSG:4326", f"EPSG:{epsg_id}", always_xy=True)

    set_plot(bounds, to_local)
    forest = place_trees(existing_forest, ways, region, to_local, tree_distance, min_distance**2)
    print(existing_forest)

    plot_trees(bounds, forest, tree_distance)
    export_map(bounds, forest, epsg_id)
    export_csv(bounds, forest, wkt_polygon, tree_distance, min_distance, epsg_id)
    export_shapefile(bounds, forest, epsg_id)


if __name__ == "__main__":
    main(WKT, EPSG_ID, TREE_DISTANCE, MIN_DISTANCE, EXISTING_TREES)
